@@ -103,7 +103,6 @@ DEFUN_ATTR(no_cs7_as, no_cs7_as_cmd,
 	return CMD_SUCCESS;
 }
 
-/* TODO: routing-key */
 DEFUN_ATTR(as_asp, as_asp_cmd,
 	   "asp NAME",
 	   "Specify that a given ASP is part of this AS\n"
@@ -300,12 +299,14 @@ static int _rout_key(struct vty *vty,
 
 	/* When libosmo-sigtran is used in ASP role, the VTY routing table node
 	 * (config-cs7-rt) is not available. However, when we add a routing key
-	 * to an AS we still have to put a matching route into the routing
-	 * table. This is done automatically by first removing the old route
+	 * to an AS we still have to put a matching dynamic route into the routing
+	 * table. This is done automatically by first removing the old dynamic route
 	 * (users may change the routing key via VTY during runtime) and then
-	 * putting a new route (see below). */
+	 * putting a new dynamic route (see below). */
 	if (cs7_role == CS7_ROLE_ASP) {
-		rt = ss7_route_table_find_route_by_dpc_mask(as->inst->rtable_system, rkey->pc, 0xffffff);
+		rt = ss7_route_table_find_route_by_dpc_mask(as->inst->rtable_system,
+							    rkey->pc, 0xffffff,
+							    true);
 		if (rt)
 			ss7_route_destroy(rt);
 	}
@@ -316,9 +317,9 @@ static int _rout_key(struct vty *vty,
 	rkey->si = si ? get_string_value(mtp_si_vals, si) : 0;	/* FIXME: input validation */
 	rkey->ssn = ssn ? atoi(ssn) : 0;			/* FIXME: input validation */
 
-	/* automatically add new route (see also comment above) */
+	/* automatically add new dynamic route (see also comment above) */
 	if (cs7_role == CS7_ROLE_ASP) {
-		if (!ss7_route_create(as->inst->rtable_system, rkey->pc, 0xffffff, as->cfg.name)) {
+		if (!ss7_route_create(as->inst->rtable_system, rkey->pc, 0xffffff, true, as->cfg.name)) {
 			vty_out(vty, "Cannot create route (pc=%s, linkset=%s) to AS %s", dpc, as->cfg.name, VTY_NEWLINE);
 			return CMD_WARNING;
 		}
@@ -436,15 +437,15 @@ void ss7_vty_write_one_as(struct vty *vty, struct osmo_ss7_as *as, bool show_dyn
 		vty_out(vty, "  asp %s%s", asp->cfg.name, VTY_NEWLINE);
 	}
 	if (as->cfg.mode_set_by_vty) {
-		vty_out(vty, "  traffic-mode %s%s", osmo_ss7_as_traffic_mode_name(as->cfg.mode), VTY_NEWLINE);
+		vty_out(vty, "  traffic-mode %s", osmo_ss7_as_traffic_mode_name(as->cfg.mode));
 		if (as->cfg.mode == OSMO_SS7_AS_TMOD_LOADSHARE) {
 			if (as->cfg.loadshare.opc_sls) {
 				vty_out(vty, " bindings sls opc-sls");
 				if (as->cfg.loadshare.opc_shift != 0)
 					vty_out(vty, " opc-shift %u", as->cfg.loadshare.opc_shift);
 			}
-			vty_out(vty, "%s", VTY_NEWLINE);
 		}
+		vty_out(vty, "%s", VTY_NEWLINE);
 
 		if (as->cfg.loadshare.sls_shift != 0)
 			vty_out(vty, "  sls-shift %u%s", as->cfg.loadshare.sls_shift, VTY_NEWLINE);
