@@ -398,7 +398,7 @@ DEFUN_ATTR(cs7_rt_upd, cs7_rt_upd_cmd,
 		return CMD_WARNING;
 	}
 
-	rt = ss7_route_alloc(rtable, dpc, mask);
+	rt = ss7_route_alloc(rtable, dpc, mask, false);
 	if (!rt) {
 		vty_out(vty, "%% Cannot allocate new route%s", VTY_NEWLINE);
 		return CMD_WARNING;
@@ -464,7 +464,7 @@ DEFUN_ATTR(cs7_rt_rem, cs7_rt_rem_cmd,
 		return CMD_WARNING;
 	}
 
-	rt = ss7_route_table_find_route_by_dpc_mask(rtable, dpc, mask);
+	rt = ss7_route_table_find_route_by_dpc_mask(rtable, dpc, mask, false);
 	if (!rt) {
 		vty_out(vty, "cannot find route to be deleted%s", VTY_NEWLINE);
 		return CMD_WARNING;
@@ -484,6 +484,8 @@ static void write_one_rtable(struct vty *vty, struct osmo_ss7_route_table *rtabl
 		vty_out(vty, "  description %s%s", rtable->cfg.description, VTY_NEWLINE);
 	llist_for_each_entry(clset, &rtable->combined_linksets, list) {
 		llist_for_each_entry(rt, &clset->routes, list) {
+			if (rt->cfg.dyn_allocated)
+				continue;
 			vty_out(vty, "  update route %s %s linkset %s",
 				osmo_ss7_pointcode_print(rtable->inst, rt->cfg.pc),
 				osmo_ss7_pointcode_print2(rtable->inst, rt->cfg.mask),
@@ -515,9 +517,13 @@ static void vty_dump_rtable(struct vty *vty, struct osmo_ss7_route_table *rtbl,
 
 		llist_for_each_entry(rt, &clset->routes, list) {
 			bool rt_avail = ss7_route_is_available(rt);
-
-			vty_out(vty, "%-16s %-5s %c %c %u %-19s %-7s %-7s %-7s%s",
-				osmo_ss7_route_print(rt),
+			bool first_rt_in_clset = (rt == llist_first_entry(&clset->routes, struct osmo_ss7_route, list));
+			/* Print route str only in first rt in combined linkset.
+			 * This allows users to easily determine visually combined
+			 * linksets: */
+			const char *rt_str = first_rt_in_clset ? osmo_ss7_route_print(rt) : "";
+			vty_out(vty, "%-16s %-5s %c %c %u %-19s %-7s %-7s %-7s %-3s%s",
+				rt_str,
 				rt_avail ? "acces" : "INACC",
 				' ',
 				'0' + rt->cfg.qos_class,
@@ -526,6 +532,7 @@ static void vty_dump_rtable(struct vty *vty, struct osmo_ss7_route_table *rtbl,
 				rt_avail ? "avail" : "UNAVAIL",
 				"?",
 				rt_avail ? "avail" : "UNAVAIL",
+				rt->cfg.dyn_allocated ? "dyn" : "",
 				VTY_NEWLINE);
 		}
 	}
@@ -727,15 +734,6 @@ DEFUN(cs7_asp_disconnect, cs7_asp_disconnect_cmd,
  * SCCP addressbook handling
  ***********************************************************************/
 
-/* SCCP addressbook */
-struct osmo_sccp_addr_entry {
-	struct llist_head list;
-	struct llist_head list_global;
-	struct osmo_ss7_instance *inst;
-	char name[32];
-	struct osmo_sccp_addr addr;
-};
-
 static struct cmd_node sccpaddr_node = {
 	L_CS7_SCCPADDR_NODE,
 	"%s(config-cs7-sccpaddr)# ",
@@ -748,101 +746,6 @@ static struct cmd_node sccpaddr_gt_node = {
 	1,
 };
 
-/* A global list that holds all addressbook entries at once
- * (see also .cfg in struct osmo_ss7_instance) */
-LLIST_HEAD(sccp_address_book_global);
-
-/* Pick an SCCP address entry from the addressbook list by its name */
-static struct osmo_sccp_addr_entry
-*addr_entry_by_name_local(const char *name,
-			  const struct osmo_ss7_instance *inst)
-{
-	struct osmo_sccp_addr_entry *entry;
-
-	llist_for_each_entry(entry, &inst->cfg.sccp_address_book, list) {
-		if (strcmp(entry->name, name) == 0) {
-			OSMO_ASSERT(entry->inst == inst);
-			return entry;
-		}
-	}
-
-	return NULL;
-}
-
-/* Pick an SCCP address entry from the global addressbook
- * list by its name */
-static struct osmo_sccp_addr_entry
-*addr_entry_by_name_global(const char *name)
-{
-	struct osmo_sccp_addr_entry *entry;
-
-	llist_for_each_entry(entry, &sccp_address_book_global,
-			     list_global) {
-		if (strcmp(entry->name, name) == 0)
-			return entry;
-	}
-
-	return NULL;
-}
-
-/*! \brief Lookup an SCCP address from the addressbook by its name.
- *  \param[out] dest_addr pointer to output the resulting sccp-address;
- *		(set to NULL if not interested)
- *  \param[in] name of the address to lookup
- *  \returns SS7 instance; NULL on error */
-struct osmo_ss7_instance *
-osmo_sccp_addr_by_name(struct osmo_sccp_addr *dest_addr,
-		       const char *name)
-{
-	struct osmo_sccp_addr_entry *entry;
-
-	entry = addr_entry_by_name_global(name);
-	if (!entry)
-		return NULL;
-
-	if (dest_addr)
-		*dest_addr = entry->addr;
-
-	return entry->inst;
-}
-
-/*! \brief Lookup an SCCP address from the addressbook of a specific instance
- *	   by its name.
- *  \param[out] dest_addr pointer to output the resulting sccp-address;
- *		(set to NULL if not interested)
- *  \param[in] name of the address to lookup
- *  \param[in] inst ss7 instance of which the address book will be searched
- *  \returns 0 on success; <0 on error */
-int osmo_sccp_addr_by_name_local(struct osmo_sccp_addr *dest_addr, const char *name,
-				 const struct osmo_ss7_instance *inst)
-{
-	struct osmo_sccp_addr_entry *entry;
-
-	entry = addr_entry_by_name_local(name, inst);
-	if (!entry)
-		return -ENOENT;
-
-	if (dest_addr)
-		*dest_addr = entry->addr;
-
-	return 0;
-}
-
-/*! \brief Reverse lookup the lookup-name of a specified SCCP address.
- *  \param[in] name of the address to lookup
- *  \returns char pointer to the lookup-name; NULL on error */
-const char *osmo_sccp_name_by_addr(const struct osmo_sccp_addr *addr)
-{
-	struct osmo_sccp_addr_entry *entry;
-
-	llist_for_each_entry(entry, &sccp_address_book_global, list_global) {
-		if (memcmp(&entry->addr, addr, sizeof(*addr)) == 0)
-			return entry->name;
-	}
-
-	return NULL;
-}
-
 /* Generate VTY configuration file snippet */
 static void write_sccp_addressbook(struct vty *vty,
 				   const struct osmo_ss7_instance *inst)
