@@ -28,6 +28,8 @@
 #include <osmocom/sigtran/osmo_ss7.h>
 #include <osmocom/sigtran/protocol/m3ua.h>
 
+#include "ss7_as.h"
+#include "ss7_asp.h"
 #include "ss7_internal.h"
 #include "ss7_route.h"
 #include "ss7_route_table.h"
@@ -219,6 +221,14 @@ static int handle_rkey_reg(struct osmo_ss7_asp *asp, struct xua_msg *inner,
 	as = osmo_ss7_as_find_by_rctx(asp->inst, rctx);
 	if (as) {
 		LOGPASP(asp, DLSS7, LOGL_NOTICE, "RKM: Found existing AS for RCTX %u\n", rctx);
+		/* Early return before allocating stuff if no space left: */
+		if (*nas_idx >= max_nas_idx) {
+			LOGPASP(asp, DLSS7, LOGL_ERROR, "RKM: not enough room for newly assigned AS (max %u AS)\n",
+				max_nas_idx+1);
+			msgb_append_reg_res(resp, rk_id, M3UA_RKM_REG_ERR_INSUFF_RESRC, 0);
+			return -1;
+		}
+
 		if (as->cfg.routing_key.pc != dpc) {
 			LOGPASP(asp, DLSS7, LOGL_ERROR, "RKM: DPC doesn't match, rejecting AS (%u != %u)\n",
 				as->cfg.routing_key.pc, dpc);
@@ -242,6 +252,14 @@ static int handle_rkey_reg(struct osmo_ss7_asp *asp, struct xua_msg *inner,
 			as->cfg.mode_set_by_peer = true;
 		}
 	} else if (asp->inst->cfg.permit_dyn_rkm_alloc) {
+		/* Early return before allocating stuff if no space left: */
+		if (*nas_idx >= max_nas_idx) {
+			LOGPASP(asp, DLSS7, LOGL_ERROR, "RKM: not enough room for newly assigned AS (max %u AS)\n",
+				max_nas_idx+1);
+			msgb_append_reg_res(resp, rk_id, M3UA_RKM_REG_ERR_INSUFF_RESRC, 0);
+			return -1;
+		}
+
 		/* Create an AS for this routing key */
 		snprintf(namebuf, sizeof(namebuf), "as-rkm-%u", rctx);
 		as = osmo_ss7_as_find_or_create(asp->inst, namebuf, OSMO_SS7_ASP_PROT_M3UA);
@@ -261,8 +279,8 @@ static int handle_rkey_reg(struct osmo_ss7_asp *asp, struct xua_msg *inner,
 		as->cfg.routing_key.pc = dpc;
 		as->cfg.routing_key.context = rctx;
 
-		/* add route for that routing key */
-		rt = ss7_route_create(as->inst->rtable_system, dpc, 0xFFFFFF, namebuf);
+		/* add dynamic route for that routing key */
+		rt = ss7_route_create(as->inst->rtable_system, dpc, 0xFFFFFF, true, namebuf);
 		if (!rt) {
 			LOGPASP(asp, DLSS7, LOGL_ERROR, "RKM: Cannot insert route for DPC %s / as %s\n",
 				osmo_ss7_pointcode_print(asp->inst, dpc), namebuf);
@@ -270,17 +288,6 @@ static int handle_rkey_reg(struct osmo_ss7_asp *asp, struct xua_msg *inner,
 			msgb_append_reg_res(resp, rk_id, M3UA_RKM_REG_ERR_CANT_SUPP_UNQ_RT, 0);
 			return -1;
 		}
-
-		/* append to list of newly assigned as */
-		if (*nas_idx >= max_nas_idx) {
-			ss7_route_destroy(rt);
-			osmo_ss7_as_destroy(as);
-			LOGPASP(asp, DLSS7, LOGL_ERROR, "RKM: not enough room for newly assigned AS (max %u AS)\n",
-				max_nas_idx+1);
-			msgb_append_reg_res(resp, rk_id, M3UA_RKM_REG_ERR_INSUFF_RESRC, 0);
-			return -1;
-		}
-		newly_assigned_as[(*nas_idx)++] = as;
 	} else {
 		/* not permitted to create dynamic RKM entries */
 		LOGPASP(asp, DLSS7, LOGL_NOTICE, "RKM: RCTX %u not found in configuration, and "
@@ -290,8 +297,10 @@ static int handle_rkey_reg(struct osmo_ss7_asp *asp, struct xua_msg *inner,
 	}
 
 	/* Success: Add just-create AS to connected ASP + report success */
-	osmo_ss7_as_add_asp(as, asp->cfg.name);
+	ss7_as_add_asp(as, asp);
 	msgb_append_reg_res(resp, rk_id, M3UA_RKM_REG_SUCCESS, rctx);
+	/* append to list of newly assigned as */
+	newly_assigned_as[(*nas_idx)++] = as;
 	return 0;
 }
 
@@ -337,7 +346,16 @@ static int m3ua_rx_rkm_reg_req(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 		if (!as)
 			continue;
 		/* Notify AS that it has an INACTIVE ASP */
-		osmo_fsm_inst_dispatch(as->fi, XUA_ASPAS_ASP_INACTIVE_IND, asp);
+		/* RFC4666 4.3.4.5: "When an ASP moves from ASP-DOWN to ASP-INACTIVE within a
+		* particular AS, a Notify message SHOULD be sent, by the ASP-UP receptor,
+		* after sending the ASP-UP-ACK, in order to inform the ASP of the current AS
+		* state."
+		*/
+		struct xua_as_event_asp_inactive_ind_pars pars = {
+			.asp = asp,
+			.asp_requires_notify = true,
+		};
+		osmo_fsm_inst_dispatch(as->fi, XUA_ASPAS_ASP_INACTIVE_IND, &pars);
 	}
 
 	return 0;
@@ -375,7 +393,9 @@ static int handle_rkey_dereg(struct osmo_ss7_asp *asp, uint32_t rctx,
 		return -1;
 	}
 
-	rt = ss7_route_table_find_route_by_dpc(inst->rtable_system, as->cfg.routing_key.pc);
+	rt = ss7_route_table_find_route_by_dpc_mask(inst->rtable_system,
+						    as->cfg.routing_key.pc, 0xffffff,
+						    true);
 	if (!rt) {
 		msgb_append_dereg_res(resp, M3UA_RKM_DEREG_ERR_UNKNOWN, 0);
 		return -1;
@@ -530,7 +550,8 @@ static int m3ua_rx_rkm_dereg_rsp(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 	return 0;
 }
 
-/* process an incoming RKM message in xua format */
+/* process an incoming RKM message in xua format
+ * This function takes ownership of xua msg passed to it. */
 int m3ua_rx_rkm(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 {
 	int rc;
@@ -560,6 +581,8 @@ int m3ua_rx_rkm(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 		rc = -1;
 		break;
 	}
+
+	xua_msg_free(xua);
 	return rc;
 }
 
@@ -599,10 +622,11 @@ void xua_rkm_cleanup_dyn_as_for_asp(struct osmo_ss7_asp *asp)
 	llist_for_each_entry_safe(as, as2, &inst->as_list, list) {
 		if (!osmo_ss7_as_has_asp(as, asp))
 			continue;
-		/* FIXME: check if there are no other ASPs! */
 		if (!as->rkm_dyn_allocated)
 			continue;
 
-		osmo_ss7_as_destroy(as);
+		/* If there are no other ASPs, destroy the AS: */
+		if (osmo_ss7_as_count_asp(as) == 1)
+			osmo_ss7_as_destroy(as);
 	}
 }
